/**
 * @file xt_dvrt_print.c This file implements the print function of Deverto's Tequet SW
 * copyright 2000-2013 Deverto ltd
 */
#include <linux/ip.h>
#include <linux/udp.h>

#include <linux/module.h>
#include <linux/skbuff.h>
#include <linux/version.h>
#include <linux/netfilter/x_tables.h>
#include <linux/netfilter/nf_conntrack_common.h>

#include <net/ip.h>
#include <net/netfilter/nf_conntrack.h>

MODULE_AUTHOR("Deverto Ltd <admin@deverto.com>");
MODULE_DESCRIPTION("Deverto TSS kernel print");
MODULE_LICENSE("GPL");
MODULE_ALIAS("ipt_dvrt_print");
#define CREATE_IP_BYTES(ip) ip & 0xff, (ip >> 8) & 0xff, (ip >> 16) & 0xff, (ip >> 24) & 0xff


static unsigned int
dvrt_print_tg4(struct sk_buff *skb, const struct xt_action_param *par)
{
    const struct iphdr *iph;
    const struct udphdr *udph;

    enum ip_conntrack_info ctinfo;
    struct nf_conn *ct;

    if (skb->protocol != htons(ETH_P_IP))
	return XT_CONTINUE;

    iph = ip_hdr(skb);
    if (iph->version != 4 )
	return XT_CONTINUE;

    if (iph->protocol != IPPROTO_UDP)
	return XT_CONTINUE;

    udph = (struct udphdr*)((void*)iph + ip_hdrlen(skb));

    ct = nf_ct_get(skb, &ctinfo);

    if (ct == NULL)
	return XT_CONTINUE;

    if (ct->mark == 0)
	return XT_CONTINUE;

    printk("skb->dev->name: %s mark: %u --- src:%d.%d.%d.%d:%d  dst:%d.%d.%d.%d:%d\n",
	   skb->dev->name,
	   skb->mark,
	   CREATE_IP_BYTES(iph->saddr), ntohs(udph->source),
	   CREATE_IP_BYTES(iph->daddr), ntohs(udph->dest));

    return XT_CONTINUE;
}

static struct xt_target dvrt_print_tg_reg __read_mostly = {
    .name       = "dvrt_print",
    .revision   = 1,
    .family     = NFPROTO_IPV4,
    .hooks      = 1 << NF_INET_POST_ROUTING,
    .table      = "mangle",
    .target     = dvrt_print_tg4,
    .targetsize = 0,
    .me         = THIS_MODULE,
};

static int __init dvrt_print_tg_init(void)
{
    printk("dvrt_print module is loaded");
    return xt_register_target(&dvrt_print_tg_reg);
}

static void __exit dvrt_print_tg_exit(void)
{
    printk("dvrt_print module is unloaded");
    xt_unregister_target(&dvrt_print_tg_reg);
}

module_init(dvrt_print_tg_init);
module_exit(dvrt_print_tg_exit);
