/**
 * @file xt_dvrt_proxy.c This file implements the proxy function of Deverto's Tequet SW
 * copyright 2000-2013 Deverto ltd
 */
#include <linux/ip.h>
#include <linux/udp.h>

#include <linux/module.h>
#include <linux/skbuff.h>
#include <linux/version.h>
#include <linux/netfilter/x_tables.h>
#include <linux/netfilter/nf_conntrack_common.h>

#include <net/ip.h>
#include <net/netfilter/nf_conntrack.h>

MODULE_AUTHOR("Deverto Ltd <admin@deverto.com>");
MODULE_DESCRIPTION("Deverto TSS kernel proxy");
MODULE_LICENSE("GPL");
MODULE_ALIAS("ipt_dvrt_proxy");

#define CREATE_IP_BYTES(ip) ip & 0xff, (ip >> 8) & 0xff, (ip >> 16) & 0xff, (ip >> 24) & 0xff

/* static unsigned int counter = 0; */

static unsigned int
dvrt_proxy_tg4(struct sk_buff *skb, const struct xt_action_param *par)
{
	const struct iphdr *iph;
	const struct udphdr *udph;

	enum ip_conntrack_info ctinfo;
	struct nf_conn *ct;

	if (skb->protocol != htons(ETH_P_IP))
	    return XT_CONTINUE;

	iph = ip_hdr(skb);
	if (iph->version != 4 )
	    return XT_CONTINUE;

	if (iph->protocol != IPPROTO_UDP)
	    return XT_CONTINUE;

	udph = (struct udphdr*)((void*)iph + ip_hdrlen(skb));

	ct = nf_ct_get(skb, &ctinfo);

	if (ct == NULL)
	    return XT_CONTINUE;

	if (ct->mark == 0)
	    return XT_CONTINUE;


	if (ct->tuplehash[IP_CT_DIR_ORIGINAL].tuple.dst.u3.ip == iph->daddr &&
	    ct->tuplehash[IP_CT_DIR_ORIGINAL].tuple.dst.u.udp.port == udph->dest)
	{
	    skb->mark = ct->mark & 0xffff;
	}
	else
	{
	    skb->mark = (ct->mark >> 16) & 0x7fff;
	}

	skb->mark |= 1 << 17;
	/*
	if ((counter % 200) == 0 || (counter % 200) == 1)
	{
	    printk("skb->dev->name: %s\n", skb->dev->name);
	    printk("conntrack entry: orig: (%d.%d.%d.%d:%d , %d.%d.%d.%d:%d), "
		   "repl: (%d.%d.%d.%d:%d , %d.%d.%d.%d:%d), upper mark: %d, bottom mark: %d\n",
	    	   CREATE_IP_BYTES(ct->tuplehash[IP_CT_DIR_ORIGINAL].tuple.src.u3.ip),
		   htons(ct->tuplehash[IP_CT_DIR_ORIGINAL].tuple.src.u.udp.port),
		   CREATE_IP_BYTES(ct->tuplehash[IP_CT_DIR_ORIGINAL].tuple.dst.u3.ip),
		   htons(ct->tuplehash[IP_CT_DIR_ORIGINAL].tuple.dst.u.udp.port),
	    	   CREATE_IP_BYTES(ct->tuplehash[IP_CT_DIR_REPLY].tuple.src.u3.ip),
		   htons(ct->tuplehash[IP_CT_DIR_REPLY].tuple.src.u.udp.port),
		   CREATE_IP_BYTES(ct->tuplehash[IP_CT_DIR_REPLY].tuple.dst.u3.ip),
		   htons(ct->tuplehash[IP_CT_DIR_REPLY].tuple.dst.u.udp.port),
		   (ct->mark >> 16) & 0x7fff, ct->mark & 0xffff);
	    printk("Packet will be routed by mark: %d, src: %d.%d.%d.%d:%d dst:%d.%d.%d.%d:%d\n",
		   skb->mark,
		   CREATE_IP_BYTES(iph->saddr), htons(udph->source),
		   CREATE_IP_BYTES(iph->daddr), htons(udph->dest));

	}

	++counter;
	*/
	return XT_CONTINUE;

}


static struct xt_target dvrt_proxy_tg_reg __read_mostly = {
    .name       = "dvrt_proxy",
    .revision   = 1,
    .family     = NFPROTO_IPV4,
    .hooks      = 1 << NF_INET_PRE_ROUTING,
    .table      = "mangle",
    .target     = dvrt_proxy_tg4,
    .targetsize = 0,
    .me         = THIS_MODULE,
};

static int __init dvrt_proxy_tg_init(void)
{
    printk("dvrt_proxy module is loaded");
    return xt_register_target(&dvrt_proxy_tg_reg);
}

static void __exit dvrt_proxy_tg_exit(void)
{
    xt_unregister_target(&dvrt_proxy_tg_reg);
    printk("dvrt_proxy module is unloaded");
}

module_init(dvrt_proxy_tg_init);
module_exit(dvrt_proxy_tg_exit);
